#-----------------------------------------------------------------------
#
#     Quest for the Holy Grail
#
#-----------------------------------------------------------------------

VERSION = "1.3"

debug = False

#-----------------------------------------------------------------------

import math, re, sys, textwrap
from getopt import getopt, GetoptError
from random import randrange, choice, shuffle, seed
from binascii import a2b_base64
from cStringIO import StringIO
try:
	set
except NameError:
	from sets import Set as set
import pygame
from pygame import Rect, Surface
from pygame.draw import arc, polygon
from pygame.color import Color
from pygame.locals import QUIT, KEYDOWN, KEYUP, USEREVENT, FULLSCREEN
try:
	data
except NameError:
	data = {}

#-----------------------------------------------------------------------

pygame.font.init()
fontname = pygame.font.get_default_font()

def get_font(size, italic = False):
	return pygame.font.Font(fontname, size, italic = italic)

#-----------------------------------------------------------------------

def decode((size, data)):
	blocks = []
	r9 = xrange(9)
	for i in xrange(0, len(data), 11):
		chars = data[i:i+11:]
		chars = chars[::-1]
		n = 0L
		for c in chars:
			n = n * 95 + (ord(c) - 32)
		buf = []
		for j in r9:
			buf.insert(0, chr(n & 0xff))
			n >>= 8
		blocks.append("".join(buf))
	return "".join(blocks)[:size]

#-----------------------------------------------------------------------

def fatal(message):
	sys.stderr.write("%s\n" % message)
	sys.exit(1)

usage = """
Usage:
   python quest.py [options] [<width>x<height>]

Options:
   -f        Full screen
   -w        Windowed
   -s size   Level size: 1, 2 or 3
   -i        Skip instructions screen
"""

class Options:
	seed = None
	start_level = 0
	show_enemy_health = False
	small_dungeon = False
	desktop_size = None
	fullscreen = False
	level_size = None
	skip_intro = False
	debug = False
	
	def __init__(self):
		args = sys.argv[1:]
		try:
			optvals, params = getopt(args, "fwids:HMS:L:")
			for option, value in optvals:
				if option == "-f":
					self.fullscreen = True
				elif option == "-w":
					self.fullscreen = False
				elif option == "-i":
					self.skip_intro = True
				elif option == "-s":
					self.level_size = value
				elif option == "-d":
					self.debug = True
				elif option == "-S":
					self.seed = 1234 + int(value)
				elif option == "-L":
					self.start_level = int(value)
				elif option == "-H":
					self.show_enemy_health = True
				elif option == "-M":
					self.small_dungeon = True
			if params:
				self.set_window_size(params[0])
		except (GetoptError, ValueError):
			fatal(usage)
	
	def set_window_size(self, value):
		numbers = [int(x) for x in value.split("x")]
		if len(numbers) <> 2:
			raise ValueError
		self.desktop_size = numbers

options = Options()

if options.debug:
	debug = True

if debug:
	db = open("quest.log", "w")

#-----------------------------------------------------------------------

default_desktop_size = (1024, 768)
	
if options.small_dungeon:
	room_width = 7
	room_height = 7
	#rooms_wide = 3
	#rooms_high = 2
	room_gap = 1
else:
	room_width = 9
	room_height = 9
	#rooms_wide = 3
	#rooms_high = 3
	room_gap = 5

min_room_width = 5
min_room_height = 5

cell_width = 32
cell_height = 32

title_box_size = 600, 460
title_font_1 = get_font(96)
title_font_2 = get_font(24)
title_font_3 = get_font(48)
title_text_color = Color("Gold")
title_bg_color = Color("blue")
version_font = get_font(12)

status_width = 260
status_left_margin = 12
status_top_margin = 8
status_text_color = Color("black")
status_warning_color = Color("red")
status_bg_color = Color("wheat")
stats_font = get_font(20)
stats_position = 230
inventory_font = get_font(15)

log_lines = 4
log_font = get_font(16)
log_left_margin = 12
log_hmargin = 16
log_vmargin = 8
log_old_text_color = Color("black")
log_text_color = Color("brown")
log_bg_color = Color("wheat")

congrat_font = get_font(48)
victory_font = get_font(24)
win_text_color = Color("gold")
win_bg_color = (100, 0, 0)
rip_font = get_font(96)
rip_text_color = Color("black")
rip_bg_color = Color("light grey")
sc_font = get_font(96)
sc_text_color = Color("black")
sc_bg_color = Color("lemonchiffon1")
sc_wrap_width = 60
play_again_font = get_font(16)

frame_time = 100 # ms
repeat_delay = 3
repeat_spacing = 1

#-----------------------------------------------------------------------

instructions_box_size = title_box_size
instructions_text_color = (100, 0, 0)
instructions_bg_color = Color("lemonchiffon2")
instructions_title_font = get_font(48)
instructions_font = get_font(20)
instructions_wrap_width = 50

instructions = """
	Use the arrow keys to move around. To pick up an object,
	walk over it. To attack an enemy, try to walk into it.

	To use an item in your inventory, press the key indicated
	in (...) after the item's name. Some items may have more
	than one use, depending on the circumstances.

	Press M to see a map of the explored parts of the
	current level.

	Press ESCAPE during play to see these instructions again.
"""

#-----------------------------------------------------------------------

options_box_size = (300, 200)
options_heading_font = get_font(24)
options_font = get_font(18)
dungeon_sizes = {"1": ("Small", (3, 3)), "2": ("Medium", (4, 4)), "3": ("Large", (5, 5))}

#-----------------------------------------------------------------------

credits_box_size = 640, 300
credits_heading_font = get_font(48)
credits_role_font = get_font(20)
credits_name_font = get_font(16, italic = True)
credits_text_color = Color("gold")

credits = [
	["Concept and Programming", ["Gregory Ewing"]],
	["Testing", ["Oliver Hunt"]],
	["Attempted Testing", ["Michael JasonSmith"]],
	["Acknowledgements", ["The Monty Python Team"]],
]

#-----------------------------------------------------------------------

colors = {}

def init_colors():
	from pygame.colordict import THECOLORS
	name_to_numbers = {}
	for key in THECOLORS:
		s = re.split("([0-9]+)", key)
		if len(s) >= 2:
			name = s[0]
			number = int(s[1])
		else:
			name = s[0]
			number = 0
		numbers = name_to_numbers.setdefault(name, [])
		if number:
			numbers.append(number)
	names = name_to_numbers.keys() + ["almond", "cream", "lemon", "rose"]
	words1 = set(names)
	words1.remove("gold")
	while 1:
		words3 = set()
		for word1 in words1:
			for word2 in words1:
				if word1 <> word2:
					i = word1.find(word2)
					if i >= 0:
						word1a = word1[:i]
						word1b = word1[i+len(word2):]
						if word1a:
							words3.add(word1a)
						words3.add(word2)
						if word1b:
							words3.add(word1b)
						break
			else:
				words3.add(word1)
		if words3 == words1:
			break
		words1 = words3
	for internal_name in name_to_numbers:
		name = internal_name
		words = []
		while name:
			for word in words3:
				if name.startswith(word):
					words.append(word)
					name = name[len(word):]
					break
			else:
				words.append(name)
				break
		display_name = " ".join(words)
		numbers = name_to_numbers[internal_name]
		if numbers:
			numbers.sort()
			number = numbers[len(numbers) // 2]
			internal_name = "%s%s" % (internal_name, number)
		colors[display_name] = THECOLORS[internal_name]

#-----------------------------------------------------------------------

def init_color_names():
	global available_color_names
	available_color_names = colors.keys()
	shuffle(available_color_names)

#-----------------------------------------------------------------------

class Window(object):

	def __init__(self, max_width, max_height):
		log_height = log_lines * log_font.get_linesize() + 2 * log_vmargin
		#cells_wide = min((max_width - status_width) // cell_width, dungeon_cells_wide)
		#cells_high = min((max_height - log_height) // cell_height, dungeon_cells_high)
		cells_wide = (max_width - status_width) // cell_width
		cells_high = (max_height - log_height) // cell_height
		arena_rect = Rect(0, 0, cells_wide * cell_width, cells_high * cell_height)
		log_rect = Rect(0, arena_rect.bottom, arena_rect.width + status_width, log_height)
		status_rect = Rect(arena_rect.right, arena_rect.top, status_width, arena_rect.height)
		status_margins = status_rect.inflate(-status_left_margin, -status_top_margin)
		screen_rect = arena_rect.unionall([log_rect, status_rect])
		self.cells_wide = cells_wide
		self.cells_high = cells_high
		self.arena_rect = arena_rect
		self.log_rect = log_rect
		self.status_rect = status_rect
		self.status_margins = status_margins
		self.screen_rect = screen_rect
		self.scroll_left = 0
		self.scroll_top = 0
	
	def viewport(self, level):
		row_min = self.scroll_top
		col_min = self.scroll_left
		return (max(row_min, 0), max(col_min, 0),
			min(row_min + window.cells_high, level.dungeon_cells_high),
			min(col_min + window.cells_wide, level.dungeon_cells_wide))
	
	def reveal_rect(self, left, top, right, bottom):
		drow = dcol = 0
		vtop = self.scroll_top
		vleft = self.scroll_left
		vright = vleft + self.cells_wide
		vbottom = vtop + self.cells_high
		if left < vleft and right < vright:
			dcol = -min(vleft - left, vright - right)
		elif left > vleft and right > vright:
			dcol = min(left - vleft, right - vright)
		if top < vtop and bottom < vbottom:
			drow = -min(vtop - top, vbottom - bottom)
		elif top > vtop and bottom > vbottom:
			drow = min(top - vtop, bottom - vbottom)
		vleft += dcol
		vright += dcol
		vtop += drow
		vbottom += drow
		self.scroll_left = vleft
		self.scroll_top = vtop

#-----------------------------------------------------------------------

DIR_RIGHT = (0, 1)
DIR_DOWN = (1, 0)
DIR_LEFT = (0, -1)
DIR_UP = (-1, 0)

directions = (DIR_DOWN, DIR_RIGHT, DIR_UP, DIR_LEFT)

white = Color("white")
black = Color("black")

KEY_RIGHT = 0x113
KEY_DOWN = 0x112
KEY_LEFT = 0x114
KEY_UP = 0x111
KEY_ESCAPE = 27

deg = math.pi / 180.0

def load(image_name, colorkey = white):
	#b64 = data.get(image_name)
	#if b64:
	#	f = StringIO(a2b_base64(b64))
	b95 = data.get(image_name[:-4])
	if b95:
		f = StringIO(decode(b95))
		image = pygame.image.load(f, image_name)
	else:
		print "*** Loading image from file", image_name
		image = pygame.image.load(image_name)
	if colorkey:
		image.set_colorkey(colorkey)
	return image
	
floors = [load("gravel.png"), load("grass.png")]
walls = [load("hedge.png"), load("stone.png")]
upstairs = load("upstairs.png")
downstairs = load("downstairs.png")


def chance(percent):
	return randrange(0, 100) < percent

def format(message, arg):
	if "%" in message:
		message %= arg
	return message

def cell_to_screen_coords(row, col, w, h, map):
	row0 = window.scroll_top
	col0 = window.scroll_left
	x = col * cell_width + (cell_width - w) / 2
	y = row * cell_height + (cell_height - h) / 2
	if not map:
		x -= col0 * cell_width
		y -= row0 * cell_height
	return x, y

def draw_tile(surface, tile, row, col, map):
	if tile:
		w, h = tile.get_size()
		x, y = cell_to_screen_coords(row, col, w, h, map)
		surface.blit(tile, (x, y))

def draw_tinted_tile(surface, tile, mask, tint, row, col, map):
	buffer = Surface(tile.get_size())
	buffer.set_colorkey(white)
	buffer.fill(tint)
	buffer.blit(mask, (0, 0))
	buffer.blit(tile, (0, 0))
	draw_tile(surface, buffer, row, col, map)

def draw_text(surface, text, font, color, center):
	image = font.render(text, True, color)
	dest = image.get_rect()
	dest.center = center
	surface.blit(image, dest)

def get_key(animation = None):
	while 1:
		event = pygame.event.wait()
		type = event.type
		if type == QUIT:
				sys.exit(0)
		elif type == KEYDOWN:
			return event.unicode.upper()
		elif type == USEREVENT:
			if animation:
				animation.update()
				pygame.display.flip()

def get_yorn(animation = None):
	while 1:
		reply = get_key(animation)
		if reply == "Y":
			return True
		elif reply == "N" or reply == chr(KEY_ESCAPE):
			return False

def get_continue_or_quit(animation = None):
	while 1:
		key = get_key(animation)
		if key in "\r\n\x03":
			return
		elif key == "Q":
			sys.exit(0)
#-----------------------------------------------------------------------

class Frame(object):

	def __init__(self):
	 from pygame.transform import flip, rotate
	 self.topleft = load("frame-corner.png")
	 self.topright = flip(self.topleft, True, False)
	 self.botleft = flip(self.topleft, False, True)
	 self.botright = flip(self.topright, False, True)
	 self.top = load("frame-side.png")
	 self.bottom = flip(self.top, False, True)
	 self.left = rotate(self.top, 90)
	 self.right = flip(self.left, True, False)
	 self.tile_size = self.topleft.get_size()
	
	def draw(self, surface, rect):
		tw, th = self.tile_size
		box = rect.inflate(tw, th)
		for x in xrange(box.left + tw, box.right - tw, tw):
			surface.blit(self.top, (x, box.top))
			surface.blit(self.bottom, (x, box.bottom - th))
		for y in xrange(box.top + th, box.bottom - th, th):
			surface.blit(self.left, (box.left, y))
			surface.blit(self.right, (box.right - tw, y))
		surface.blit(self.topleft, box.topleft)
		surface.blit(self.topright, (box.right - tw, box.top))
		surface.blit(self.botleft, (box.left, box.bottom - th))
		surface.blit(self.botright, (box.right - tw, box.bottom - th))

frame = Frame()

#-----------------------------------------------------------------------

class AbortCommand(Exception):
	pass

class DungeonGenerationError(Exception):
	pass

#-----------------------------------------------------------------------

class TextCursor(object):
	surface = None
	left_margin = None
	color = None
	font = None
	tabs = None

	def __init__(self, surface, margin = 0, top = 0, color = None, font = None, just = 1):
		self.surface = surface
		self.margin = margin
		self.color = color
		self.font = font
		self.just = just
		self.x = margin
		self.y = top
	
	def write(self, text = None, font = None, color = None, smooth = True, newline = False):
		font = font or self.font
		if text:
			color = color or self.color
			image = font.render(text, smooth, color)
			width = image.get_width()
			just = self.just
			if just < 0:
				self.x -= width
				x = self.x
			elif just == 0:
				x = self.x - width // 2
			else:
				x = self.x
			self.surface.blit(image, (x, self.y))
			if just > 0:
				self.x += width
		if newline:
			self.x = self.margin
			self.y += font.get_linesize()

	def writeln(self, *args, **kwds):
		self.write(newline = True, *args, **kwds)
	
#-----------------------------------------------------------------------

class MessageLog(object):

	def __init__(self, rect):
		self.rect = rect
		self.margins = rect.inflate(-log_hmargin, -log_vmargin)
		self.clear()
	
	def clear(self):
		self.lines = []
		self.top_line = 0
		self.mark_line = 0
		#self.mark_flag = False
	
	def add_line(self, text):
		#if self.mark_flag:
		#	self.mark_line = len(self.lines)
		#	self.mark_flag = False
		self.lines.append(text)
		self.top_line = max(len(self.lines) - log_lines, self.top_line)
	
	def mark(self):
		#self.mark_flag = True
		self.mark_line = len(self.lines)
	
	def draw(self, surface):
		surface.fill(log_bg_color, self.rect)
		tc = TextCursor(surface, margin = window.log_rect.left + log_hmargin,
			top = window.log_rect.top + log_vmargin, font = log_font, color = log_text_color)
		top = self.top_line
		for i, line in enumerate(self.lines[top:top+4]):
			if top + i >= self.mark_line:
				color = log_text_color
			else:
				color = log_old_text_color
			tc.writeln(line, color = color)

def log(message):
	message_log.add_line(message)

#-----------------------------------------------------------------------

class Item(object):
	image = None
	name = None
	short_name = None
	plural_endings = ["", "s"]
	seen_message = None
	taken_message = "You find %s."
	command = None
	indefinite_article = "a"
	in_corridor = False
	on_through_path = False
	is_prerequisite = False
	is_obstacle = False
	place_last = False
	not_near_wall = False
	animated = False
	level_jitter = 0
	auxiliary_recipe = []
	group_prerequisites = False
	scale_quantities = True
	show_on_map = False
	invisible = False

	prerequisite_for = None
	prerequisites = None
	level = None
	coords = (0, 0)
	seen = False
	
	def __init__(self, prerequisite_for = None):
		self.prerequisite_for = prerequisite_for
		self.prerequisites = []

	def __str__(self):
		return "%s at %s of level %s" % (self.def_name(), self.coords, self.level)

	def get_name(self, number = 1, a = False, long = False):
		if long:
			name = self.name
		else:
			name = self.short_name or self.name
		name = format(name, self.plural_endings[number <> 1])
		if a:
			article = self.indefinite_article
			if article == "a" and name[0].lower() in "aeiou":
				article = "an"
			name = "%s %s" % (article, name)
		return name
	
	def def_name(self, long = False):
		return self.get_name(long = long)
	
	def indef_name(self, long = False):
		return self.get_name(a = True, long = long)
	
	def log(self, message):
		log(format(message, self.def_name()))
	
	def get_cell(self):
		if self.level:
			row, col = self.coords
			return self.level.get_cell(row, col)
		else:
			return None
	
	def get_room(self):
		cell = self.get_cell()
		if cell:
			return cell.room
		else:
			return None
	
	def add_to_level(self, level):
		self.level = level
		if self.animated:
			level.animated_items.append(self)
	
	def remove_from_level(self):
		level = self.level
		if self.animated:
			level.animated_items.remove(self)
		row, col = self.coords
		cell = level.get_cell(row, col)
		cell.item = None
		self.level = None
		self.coords = None

	def move_to_coords(self, row2, col2, new_level = None):
		#print "Moving", self.__class__.__name__, "to", (row2, col2), ###
		#if new_level: ###
		#	print "of level", new_level.number ###
		#else: ###
		#	print ###
		level = self.level
		if level:
			row, col = self.coords
			cell = level.get_cell(row, col)
			cell.item = None
		if new_level and new_level <> level:
			if level:
				self.remove_from_level()
			self.add_to_level(new_level)
			level = new_level
		cell = level.get_cell(row2, col2)
		if not cell.floor:
			#raise RuntimeError("Item placed outside level")
			print "Item placed outside level %d:" % level.number, self ###
		if cell.wall:
			#raise RuntimeError("Item placed in wall")
			print "Item placed in wall:", self ###
		cell.item = self
		self.coords = row2, col2

	def draw(self, surface, row, col, map):
		if not self.invisible:
			self.draw_tile(surface, row, col, map)
	
	def draw_tile(self, surface, row, col, map):
		draw_tile(surface, self.image, row, col, map)
	
	def make_seen(self):
		if not self.seen and not self.invisible:
			self.seen = True
			self.report_seen()
	
	def report_seen(self):
		if self.seen_message:
			log(format(self.seen_message, self.indef_name(long = True)))
	
	def attack(self, player):
		self.report_taken()
		row, col = self.coords
		self.remove_from_level()
		player.add_to_inventory(self)
		player.step_to_coords(row, col)
		return True
	
	def report_taken(self):
		log(format(self.taken_message, self.indef_name()))
	
#-----------------------------------------------------------------------

class TintedItem(Item):

	def draw_tile(self, surface, row, col, map):
		draw_tinted_tile(surface, self.image, self.mask, self.tint, row, col, map)
	
#-----------------------------------------------------------------------

class Actor(Item):
	health = None
	fortitude = None
	hostile = True
	mobile = True
	tracking_range = 3
	wander_chance = 10
	confined_to_room = False
	invisible = False
	floating = False
	seen_message = "You encounter %s."
	block_message = "The %s won't let you pass."
	attack_message = "The %s assaults you."
	hit_message = "You assault the %s."
	miss_message = "You swing at the %s, but miss."
	victory_messages = ["You have been defeated by %s."] * 2
	defeat_message = "The %s gives up and runs away."
	revival_message = """Fortunately, you are found by a passing shrubber,
		who uses some of the supplies in your pack to restore your health
		with a nourishing meal of spam, bacon, sausages, eggs and spam."""

	def add_to_level(self, level):
		Item.add_to_level(self, level)
		if self.floating:
			level.game.actors.add(self)
		else:
			level.actors.append(self)
	
	def remove_from_level(self):
		if not self.floating:
			self.level.actors.remove(self)
		Item.remove_from_level(self)
	
	def get_victory_message(self, fatal):
		return format(self.victory_messages[not fatal], self.indef_name(long = True))
	
	def get_revival_message(self):
		return self.revival_message
	
	def attack(self, player):
		if self.invisible:
			log("Something is blocking your way.")
		else:
			if self.hostile:
				self.take_damage(player)
			else:
				self.log(self.block_message)

	def take_damage(self, player):
		damage = self.calculate_damage_from_player(player)
		if damage is not None:
			self.report_damage_from_player(damage)
			if damage > 0:
				self.health -= damage
				if self.health <= 0:
					self.log(self.defeat_message)
					self.die()
			return damage
		else:
			return 0
	
	def die(self):
		self.remove_from_level()
	
	def calculate_damage_from_player(self, player):
		f = player.fortitude
		return randrange(f // 2, f + 1)
	
	def report_damage_from_player(self, damage):
		if damage > 0:
			self.log(self.hit_message)
		else:
			self.log(self.miss_message)
		if options.show_enemy_health:
			self.log("[%%s health is now %s]" % self.health)
		#print self.__class__.__name__, "health =", self.health ###
	
	def take_turn(self, player):
		if self.level:
			distance = self.distance_from(player)
			if self.hostile and self.in_attack_range(player, distance):
				self.reveal_to_player(player)
				self.attack_player(player)
			elif self.mobile:
				if distance <= self.tracking_range:
					self.move_towards(player)
				elif chance(self.wander_chance):
					self.try_to_move_in_direction(choice(directions))
	
	def in_attack_range(self, player, distance):
		return distance <= 1
	
	def reveal_to_player(self, player):
		self.make_seen()
	
	def attack_player(self, player):
		damage = self.calculate_damage_to_player(player)
		self.report_damage_to_player(damage)
		player.take_damage(damage, self)
	
	def calculate_damage_to_player(self, player):
		return randrange(1, self.fortitude + 1)
	
	def report_damage_to_player(self, damage):
		self.log(self.attack_message)

	def try_to_move_in_direction(self, (drow, dcol)):
		row, col = self.coords
		row2 = row + drow
		col2 = col + dcol
		cell = self.level.get_cell(row2, col2)
		if cell and not (cell.wall or cell.item or cell.stairs):
			if not self.confined_to_room or cell.room:
				self.move_to_coords(row2, col2)
	
	def move_towards(self, player):
		row, col = self.coords
		prow, pcol = player.coords
		if row < prow:
			row += 1
		elif row > prow:
			row -= 1
		if col < pcol:
			col += 1
		elif col > pcol:
			col -= 1
		cell = self.level.get_cell(row, col)
		if cell and not (cell.wall or cell.item or cell.stairs):
			self.move_to_coords(row, col)
	
	def distance_from(self, other):
		row1, col1 = self.coords
		row2, col2 = other.coords
		return abs(row1 - row2) + abs(col1 - col2)
	
	def fling(self, actor, to_level, pad = 0):
		row, col = to_level.random_position(pad)
		actor.move_to_coords(row, col, to_level)

#-----------------------------------------------------------------------

class Spam(Item):
	image = load("spam.png")
	name = "tin%s of spam"
	command = "S"
	level_jitter = 50
	
	def use(self, player):
		target = player.get_target_item()
		if isinstance(target, Shrubber):
			log("The shrubber accepts your payment and gives you a shrubbery.")
			player.add_to_inventory(Shrubbery())
		else:
			player.add_health(10)
			player.add_fortitude(5)
			log("You gobble a tin of spam.")
		player.remove_from_inventory(self)
	
#-----------------------------------------------------------------------

class MSW(Actor):
	image = load("msw.png")
	name = "Minister%s of Silly Walks"
	short_name = "minister%s"
	health = 5
	fortitude = 4
	tracking_range = 2
	hit_message = "You show the %s a silly walk of your own."
	attack_message = "The antics of the %s make you nearly laugh your head off."
	victory_messages = [
		"You have laughed yourself to death.",
		"You have laughed yourself senseless."
	]
	defeat_message = "The %s judges your walk worthy of a government grant."

	def attack(self, player):
		if chance(30):
			log("The minister's antics paralyse you with laughter.")
		else:
			damage = self.take_damage(player)
			if self.health > 0:
				if damage < 3:
					log('"It\'s not really very silly," he remarks.')

#-----------------------------------------------------------------------

class Castle(Item):
	image = load("castle.png")
	name = "abandoned castle%s"
	
	def attack(self, player):
		log("There is a run-down French castle here.")
		log("Either there's nobody home, or they're not answering the door.")

#-----------------------------------------------------------------------

class French(Actor):
	image = load("french.png")
	name = "upstart Frenchman%s"
	short_name = "Frenchman%s"
	health = 10
	fortitude = 3
	mobile = False
	not_near_wall = True
	insults = [
		"says something uncomplimentary about your mother",
		"makes a scathing remark about English kaniggets",
		"puts his thumbs in his ears and waggles his fingers at you",
		"blows a raspberry at you",
		"makes a rude gesture",
		"farts in your general direction",
	]
	responses = [
		"You pointedly ignore him.",
		"You aggressively ignore him.",
		"You ignore him really hard.",
	]
	victory_messages = [
		"You have been mortally insulted by %s.",
		"You have lost your temper with %s."
	]
	defeat_message = "The %s retreats sullenly into his castle."
	revival_message = """Fortunately, a passing shrubber intervenes
		and defuses the situation, preventing an international incident.
		But it takes two whole tins of spam to bring your blood
		pressure down to normal."""
	
	def in_attack_range(self, player, distance):
		return self.get_cell().room is player.get_cell().room
	
	def calculate_damage_to_player(self, player):
		if chance(60):
			return Actor.calculate_damage_to_player(self, player)
		else:
			return 0
	
	def report_damage_to_player(self, damage):
		if damage > 0:
			addendum = ", hurting your feelings"
		else:
			addendum = ""
		self.log("The %%s %s%s." % (choice(self.insults), addendum))
	
	def report_damage_from_player(self, damage):
		msgs = self.responses
		n = len(msgs)
		i = int((damage * n) // 11)
		log(msgs[min(i, n - 1)])
	
	def die(self):
		row, col = self.coords
		level = self.level
		Actor.die(self)
		castle = Castle()
		castle.move_to_coords(row, col, level)

#-----------------------------------------------------------------------

class Cracker(Item):
	image = load("cracker.png")
	name = "cracker%s of resurrection"
	short_name = "cracker%s"
	command = "C"
	level_jitter = 50
	
	def use(self, player):
		target = player.get_target_item()
		if isinstance(target, Parrot):
			log("The cracker revives the parrot. It seems it was only resting.")
			target.remove_from_level()
			player.remove_from_inventory(self)
			return True
		else:
			log("There's nothing here that wants a cracker.")

#-----------------------------------------------------------------------

class Parrot(Actor):
	image = load("parrot.png")
	name = "dead parrot%s"
	mobile = False
	hostile = False
	in_corridor = True
	show_on_map = True
	auxiliary_recipe = [(Cracker, 1, 1)]
	group_prerequisites = True
	
	def attack(self, player):
		self.log("The %s is blocking your path.")

#-----------------------------------------------------------------------

class Arithmetic(Item):
	image = load("arithmetic.png")
	name = "arithmetic lesson%s"
	indefinite_article = "an"
	command = "A"
	scale_quantities = False
	
	def use(self, player):
		log("Your counting ability makes a dramatic improvement.")
		player.add_intelligence(5)
		player.remove_from_inventory(self)
		return True

#-----------------------------------------------------------------------

class Spanish(Actor):
	image = load("spanish.png")
	name = "Spanish Inquisition%s"
	invisible = True
	health = 15
	fortitude = 5
	wander_chance = 0
	attack_message = "The %s assails you with an almost fanatical devotion to the Pope."
	victory_messages = [
		"You have been fatally surprised by %s.",
		"You have been surprised senseless by %s."
	]
	defeat_message = \
		'"We\'ll be back when you least expect it!" says the Inquisition as it departs.'
	
	def reveal_to_player(self, player):
		self.seen = True
		if self.invisible:
			self.invisible = False
			log("You are surprised by a Spanish Inquisition.")

	def calculate_damage_from_player(self, player):
		i = player.intelligence
		f = player.fortitude // 2
		if i > f:
			damage = randrange(i // 2, i + 1)
			if i >= 15:
				self.log("You bamboozle the %s with your brilliant counting skills.")
			else:
				self.log("You confuse the %s with your counting ability.")
		else:
			damage = randrange(f // 2, f + 1)
			message = "You fend off the %s with brute strength."
			if i <= 5:
				message += " Pity you can't count better..."
			self.log(message)
		return damage

	def report_damage_from_player(self, damage):
		pass

#-----------------------------------------------------------------------

class Bran(Item):
	image = load("bran.png")
	name = "packet%s of Pope Bran"
	command = "B"
	level_jitter = 50
	
	def use(self, player):
		player.add_moral_fibre(10)
		player.remove_from_inventory(self)
		log("You wolf a packet of Pope Bran. Tasty, and it boosts your moral fibre.")

#-----------------------------------------------------------------------

class Nun(Actor):
	image = load("nun.png")
	name = "attractive young nun%s"
	short_name = "nun%s"
	health = 5
	attack_message = "The %s plies you with dire temptation."
	victory_messages = [
		"You have been seduced by %s.",
		"You were nearly seduced by %s."
	]
	defeat_message = "The %s gives up and retreats in disappointment."
	revival_message = """Fortunately, your fellows turn up just in time
		to rescue you from a fate better -- er, worse than death. You
		need two whole tins of spam to regain your composure."""
	auxiliary_recipe = [(Bran, 1, 1)]
	group_prerequisites = True

	def calculate_damage_from_player(self, player):
		mf = player.moral_fibre
		if mf >= 5:
			return randrange(0, mf - 3)
		else:
			log("You don't have the moral fibre to resist the nun's enticements.")
			return None
	
	def report_damage_from_player(self, damage):
		if damage > 0:
			log("You valiantly fend off the nun's advancements.")
		else:
			log("Distracted by the nun's charms, you momentarily forget your resolve.")

	def attack_player(self, player):
		log("The nun plies you with dire temptation.")
		damage = randrange(0, 2)
		player.add_moral_fibre(-damage)
		if player.moral_fibre == 0:
			player.die(self)

#-----------------------------------------------------------------------

class Shrubber(Actor):
	image = load("shrubber.png")
	name = "shrubber%s"
	mobile = False
	hostile = False
	is_prerequisite = True
	place_last = True
	not_near_wall = True
	scale_quantities = False
	show_on_map = True
	times_hit = 0
	
	def attack(self, player):
		self.times_hit += 1
		if self.times_hit >= 10:
			log('"If that\'s how you feel, I\'ll go somewhere else."')
			self.fling(self, self.level)
			self.times_hit = 0
			self.seen = False
		else:
			log("Assaulting an honest shrubber would hardly be chivalrous.")
			self.introduce()
	
	def report_seen(self):
		log("You encounter a shrubber.")
		self.introduce()
	
	def introduce(self):
		log('"For my services, I require only a modest payment of spam," he says.')

#-----------------------------------------------------------------------

class Shrubbery(Item):
	name = "shrubber%s"
	plural_endings = ["y", "ies"]
	command = "G"
	
	def use(self, player):
		target = player.get_target_item()
		if isinstance(target, Knight):
			target.hostile = False
			target.shrubberies_wanted -= 1
			if target.shrubberies_wanted > 0:
				log('The knight says, "That\'s nice, but now I want... ANOTHER shrubbery!"')
			else:
				log('"What a lovely shrubbery! Thank you," says the knight, and lets you pass.')
				target.remove_from_level()
			player.remove_from_inventory(self)
		else:
			log("There's nobody here who wants a shrubbery.")
			
#-----------------------------------------------------------------------

class Knight(Actor):
	image = load("knight.png")
	name = "knight%s"
	health = 100
	fortitude = 5
	mobile = False
	in_corridor = True
	scale_quantities = False
	show_on_map = True
	attack_message = 'The %s says "Ni!" at you.'
	hit_message = "You clash swords with the %s, and wound him a little."
	block_message = '"I want another shrubbery!" says the %s.'
	victory_messages = [
		"You have been Ni-ed to death by %s.",
		"You have been Ni-ed unconscious by %s."
	]
	defeat_message = \
		"The knight succumbs to your relentless onslaught. If only there had been a better way..."
	auxiliary_recipe = [(Spam, 2, 2)]
	
	def __init__(self, *args, **kwds):
		Actor.__init__(self, *args, **kwds)
		self.shrubberies_wanted = randrange(1, 3)
	
	def report_damage_to_player(self, damage):
		if chance(25):
			log('The knight says, "If you want me to let you through, bring me a shrubbery!"')
		else:
			Actor.report_damage_to_player(self, damage)
	
	def calculate_damage_to_player(self, player):
		return 1

#-----------------------------------------------------------------------

class BlackKnight(Actor):
	image = load("black.png")
	name = "Black Knight%s"
	mobile = False
	hostile = False
	in_corridor = True
	on_through_path = True
	scale_quantities = False
	fortitude = 10
	limbs = None
	auxiliary_recipe = [(Spam, 2, 2)]
	
	def __init__(self, *args, **kwds):
		Actor.__init__(self, *args, **kwds)
		self.limbs = ["arm", "arm", "leg", "leg"]
		shuffle(self.limbs)
	
	def attack(self, player):
		f = player.fortitude - 5
		if chance(30):
			self.log("The %s parries your thrust.")
		elif randrange(6) <= f:
			limb = choice(self.limbs)
			count = self.limbs.count(limb)
			if count > 1:
				log("You lop off one of the Black Knight's %ss." % limb)
			else:
				log("You lop off the Black Knight's other %s." % limb)
			self.limbs.remove(limb)
		else:
			self.log("Your swing fails to dent the %s's armour.")
		if self.limbs:
			self.attack_player(player)
		else:
			if chance(50):
				log('"Come back here and fight like a man!" he yells as you walk past him.')
			else:
				log("He gnaws at your ankles as you step over him.")
			self.remove_from_level()
	
	def report_damage_to_player(self, damage):
		arms, legs = self.count_limbs()
		if legs == 2:
			if arms == 2:
				self.log("The %s swings at you with his two-handed sword.")
			elif arms == 1:
				self.log("The %s swings at you with his two^H^H^Hone-handed sword.")
			else:
				self.log("The %s kicks at you vicously.")
		elif legs == 1:
			if arms:
				self.log("The %s swings at you while balancing on one foot.")
			else:
				self.log("The %s hops and attacks you with a head-butt.")
		else:
			self.log("The %s takes a swipe at your knees.")
	
	def count_limbs(self):
		return self.limbs.count("arm"), self.limbs.count("leg")
	
	def get_victory_message(self, fatal):
		descs = []
		for limb, adj in [("arm", "armed"), ("leg", "legged")]:
			count = self.limbs.count(limb)
			if count < 2:
				descs.append("%s-%s" % (("no", "one")[count], adj))
		desc = ", ".join(descs) + " "
		if desc:
			desc += " "
		return "You have been %s by a %sBlack Knight." % (("clobbered", "killed")[fatal], desc)

#-----------------------------------------------------------------------

class General(Item):
	image = load("general.png")
	name = "scroll%s of general knowledge"
	command = "K"
	is_prerequisite = True
	scale_quantities = False
	level_jitter = 70
	
	def use(self, player):
		log("You learn a variety of obscure and useless facts.")
		player.add_general_knowledge(5)
		player.remove_from_inventory(self)

#-----------------------------------------------------------------------

class Hermit(Actor):
	image = load("hermit.png")
	name = "mysterious hermit%s"
	short_name = "hermit%s"
	mobile = False
	hostile = False
	in_corridor = True
	on_through_path = True
	scale_quantities = False
	auxiliary_recipe = [(General, 3, 3)]
	topics = None
	answered = 0
	
	def __init__(self, *args, **kwds):
		Actor.__init__(self, *args, **kwds)
		self.reset()
	
	def reset(self):
		topics = ["geography", "history", "physics", "chemistry", "biology", "calculus"]
		shuffle(topics)
		self.topics = topics[:2] + ["swallows"]
		self.answered = 0
		self.seen = 0
	
	def report_seen(self):
		Actor.report_seen(self)
		log('"Three questions shall ye answer before ye shall pass!" cackles the hermit.')
	
	def attack(self, player):
		if self.answered < 3:
			topic = self.topics[self.answered]
			log("The hermit asks you an obscure question about %s." % topic)
			#if randrange(15) < player.general_knowledge:
			if player.general_knowledge // 5 > self.answered:
				self.answered += 1
				if self.answered < 3:
					log("You call upon your general knowledge and answer the question correctly.")
				else:
					log('"African swallow or European swallow?" you ask.')
					log('"I don\'t know that!" he replies, and is flung to a distant part of the dungeon.')
					self.fling(self, self.level, pad = 1)
			else:
				log('"I don\'t know that!" you reply, and are flung to a distant part of the dungeon.')
				self.fling(player, self.level.level_above)
				self.reset()
		else:
			log('"Bugger off!" says the hermit.')
	
#-----------------------------------------------------------------------

class Grail(Item):
	name = "Holy Grail%s"
	animated = True
	points = [(0, 0), (0, 2), (1, 5), (3, 8), (5, 10), (8, 12), (8, 20), (0, 22), (0, 24)]
	base_color = Color("gold")
	delta_color = 0
	delta_delta_color = 7
	max_delta_color = 40
	color = base_color
	taken_message = "You've found the grail! Now all you have to do is get out of here alive..."
	
	def attack(self, player):
		Item.attack(self, player)
		player.has_grail = True
		player.game.repopulate_dungeon()
	
	def draw(self, surface, row, col, map):
		x, y = cell_to_screen_coords(row, col, 0, 0, map)
		self.draw_at(surface, x, y, 1)
	
	def draw_at(self, surface, x, y, scale):
		coords1 = []
		coords2 = []
		for px, py in self.points:
			coords1.append((x + (px - 12) * scale, y + (py - 12) * scale))
			coords2.append((x + (12 - px) * scale, y + (py - 12) * scale))
		coords = coords1 + coords2[::-1]
		thick = (scale + 1) / 2
		polygon(surface, self.color, coords)
		polygon(surface, black, coords, thick)
	
	def do_animation(self):
		d = self.delta_color
		self.color = tuple([min(255, c + d) for c in self.base_color])
		d += self.delta_delta_color
		self.delta_color = d
		if d >= self.max_delta_color or d <= 0:
			self.delta_delta_color *= -1

#-----------------------------------------------------------------------

class Bunny(Actor):
	image = load("bunny.png")
	name = "vorpal bunny%s"
	health = 50
	fortitude = 7
	mobile = True
	tracking_range = 7
	hostile = False
	seen_message = "You come across a cute little bunny rabbit."
	hit_message = "You strike desperately at the %s."
	attack_message = "The %s attacks you viciously."
	victory_messages = [
		"You have been mauled to death by %s.",
		"You have passed out from loss of blood."
	]
	defeat_message = "The %s seems to be dead (let's hope so, anyway)."
	auxiliary_recipe = [(Spam, -2, 1)]
	attacked = False
	
	def attack(self, player):
		if not self.hostile:
			log("You reach out to give the bunny a pat.")
			self.hostile = True
		else:
			Actor.attack(self, player)
	
	def report_damage_to_player(self, damage):
		if not self.attacked:
			log("It's a vorpal bunny! It attacks you vicously.")
			self.attacked = True
		else:
			Actor.report_damage_to_player(self, damage)
	
	def take_turn(self, player):
		Actor.take_turn(self, player)
		if self.seen:
			self.hostile = True

#-----------------------------------------------------------------------

class Key(TintedItem):
	image = load("key.png")
	mask = load("key_mask.png", black)
	level_jitter = 50
	
	def __init__(self, prerequisite_for):
		Item.__init__(self, prerequisite_for)
		self.name = "%s key" % prerequisite_for.color_name
		self.tint = prerequisite_for.tint
		prerequisite_for.key = self

#-----------------------------------------------------------------------

class Door(TintedItem):
	image = load("door.png")
	mask = load("door_mask.png", black)
	name = "door%s"
	in_corridor = True
	scale_quantities = True
	show_on_map = True
	seen_message = "You encounter %s."
	auxiliary_recipe = [(Key, 1, 1)]
	
	def __init__(self, *args, **kwds):
		Item.__init__(self, *args, **kwds)
		self.color_name = available_color_names.pop()
		self.name = "%s door" % self.color_name
		self.tint = colors[self.color_name]
	
	def attack(self, player):
		if self.key in player.inventory:
			log("Fortunately you have a key for the %s." % self.def_name())
			self.remove_from_level()
			player.remove_from_inventory(self.key)
		else:
			self.log("You don't have a key for the %s.")

#-----------------------------------------------------------------------

class Swallows(Actor):
	image = load("swallows.png")
	name = "pair%s of swallows"
	hostile = False
	mobile = True
	wander_chance = 80
	confined_to_room = True
	tracking_range = 0
	level_jitter = 10
	floating = True
	scale_quantities = False
	coconut_delay_range = (100, 300)
	coconut_turns = None
	seen_message = "A pair of swallows flitter by, without so much as a cord of line between them."
	
	def attack(self, player):
		log("The swallows easily avoid you.")
	
	def take_turn(self, player):
		#print "Swallows.take_turn" ###
		Actor.take_turn(self, player)
		t = self.coconut_turns
		if t is not None and t > 0:
			t -= 1
		if t == 0:
			#print "Attempting to place coconut" ###
			prow, pcol = player.coords
			cell = player.level.get_cell(prow, pcol)
			room = cell.room
			#print "Player is in room", room ###
			if room:
				for z in xrange(10):
					s_coords = player.level.random_position_in_room(room)
					c_coords = player.level.random_position_in_room(room)
					if s_coords and c_coords and s_coords <> c_coords:
						log("A pair of swallows fly into the room and drop a coconut.")
						row, col = s_coords
						self.move_to_coords(row, col, player.level)
						row, col = c_coords
						item = Coconut()
						item.move_to_coords(row, col, player.level)
						t = None
						break
		self.coconut_turns = t

#-----------------------------------------------------------------------

class Cord(Item):
	image = load("cord.png")
	name = "cord%s of line"
	command = "L"
	level_jitter = 10
	scale_quantities = False
	
	def use(self, player):
		target = player.get_target_item()
		if isinstance(target, Swallows):
			self.log("The swallows take the %s and fly off to hunt coconuts.")
			target.coconut_turns = randrange(*target.coconut_delay_range)
			player.remove_from_inventory(self)
			#target.fling(target, target.level)
			target.remove_from_level()
		else:
			self.log("There's nobody there who seems to want a cord of line.")

#-----------------------------------------------------------------------

class Coconut(Item):
	image = load("coconut.png")
	name = "coconut%s"
	command = "N"
	
	def use(self, player):
		log("You munch a coconut. Oodles of body-building nutrients!")
		player.health += 5
		player.max_health += 5
		player.base_fortitude += 5
		player.fortitude = min(player.fortitude + 5, player.base_fortitude)
		player.remove_from_inventory(self)

#-----------------------------------------------------------------------

class Grenade(Item):
	image = load("grenade.png")
	name = "Holy Hand Grenade%s"
	command = "H"

	def use(self, player):
		log("One, two, FIVE!")
		kills = 0
		prow, pcol = player.coords
		for row in xrange(prow - 1, prow + 2):
			for col in xrange(pcol - 1, pcol + 2):
				if row <> col:
					cell = player.level.get_cell(row, col)
					if cell:
						item = cell.item
						if isinstance(item, Actor):
							item.remove_from_level()
							kills += 1
		if kills:
			log("Well, that takes care of them.")
		else:
			log("There was nobody in range.")
		player.remove_from_inventory(self)

#-----------------------------------------------------------------------

class GrailAnimation(object):

	def __init__(self, surface, x, y, scale):
		self.surface = surface
		self.x = x
		self.y = y
		self.scale = scale
		self.grail = Grail()
	
	def update(self):
		self.grail.do_animation()
		self.grail.draw_at(self.surface, self.x, self.y, self.scale)

#-----------------------------------------------------------------------

if options.small_dungeon:
	from small_levels import level_recipes
else:
	level_recipes = [
		[(MSW, 3, 7),  (French, 1, 1),                                                                                                            (Spam, 3, 5), (Door, 0, 2)],  #  Level 0
		[(MSW, 5, 10), (Parrot, 1, 2),      (French, 0, 1),                                                                                       (Spam, 1, 3), (Door, 0, 2)],  #  Level 1
		[(MSW, 0, 8),  (Parrot, 1, 3),      (Spanish, 2, 4), (French, -1, 1),                                                    (Bran, 1, 3),    (Spam, 1, 2), (Door, 0, 2)],  #  Level 2
		[(MSW, 0, 6),  (Parrot, 0, 1),      (Spanish, 3, 7), (Nun, 2, 4), (Swallows, 1, 1), (Arithmetic, 2, 3), (Cord, 1, 1),                     (Spam, 1, 2), (Door, 0, 2)],  #  Level 3
		[(MSW, 0, 5),  (Parrot, -1, 1),     (Spanish, 1, 3), (Nun, 3, 7), (Knight, 1, 2),   (Shrubber, 1, 1),                    (Grenade, 0, 1), (Spam, 1, 2), (Door, 0, 2)],  #  Level 4
		[(MSW, 0, 4),  (BlackKnight, 1, 1), (Spanish, 0, 3), (Nun, 2, 5), (Knight, 1, 2),   (Swallows, 0, 1),   (Cord, 1, 1),                     (Spam, 1, 2), (Door, 0, 2)],  #  Level 5
		[(MSW, 0, 3),  (Hermit, 1, 1),      (Spanish, 0, 2), (Nun, 1, 4), (Knight, 0, 1),   (Grenade, -2, 1),                                     (Spam, 1, 2), (Door, 0, 2)],  #  Level 6
		[(MSW, 0, 5),  (Bunny, 3, 6),       (Spanish, 0, 1), (Nun, 0, 3),                   (Grenade, -4, 1),   (Cord, -2, 1),                    (Spam, 1, 2), (Door, 0, 2)],  #  Level 7
	]

repopulation_recipe = [
	(MSW, -3, 10), (Spanish, -3, 8), (Spanish, -3, 8), (Nun, -3, 5), (Bunny, 3, 7), (Bunny, 3, 7), (Bunny, 3, 7), (Spam, 0, 2)
]

#-----------------------------------------------------------------------

class Cell(object):
	floor = None
	stairs = None
	item = None
	wall = None
	room = None
	revealed = options.small_dungeon
	
	def draw(self, surface, row, col, map = False):
		if self.revealed:
			draw_tile(surface, self.floor, row, col, map)
			if self.stairs:
				draw_tile(surface, self.stairs, row, col, map)
			item = self.item
			if item and (not map or item.show_on_map):
				item.draw(surface, row, col, map)
			draw_tile(surface, self.wall, row, col, map)

#-----------------------------------------------------------------------

class Room(object):
	level = None
	coords = None
	exits = None # (drrow, drcol) -> (row, col)
	#corridors = None # (drrow, drcol) -> corridor
	corridors = None # list of Corridor
	available_positions = None # list of cell coords
	connected = False
	#is_populated = False
	#is_storeroom = False
	contains_obstacle = False
	revealed = False
	distance_from_start = None

	def __init__(self, cells, rrow, rcol, level):
		self.level = level
		self.coords = rrow, rcol
		self.exits = {}
		#self.corridors = {}
		self.corridors = []
		#self.populated_with = set()
		floor = choice(floors)
		wall = choice(walls)
		width = randrange(min_room_width, room_width + 1)
		height = randrange(min_room_height, room_height + 1)
		left = rcol * (room_width + room_gap) + randrange(room_width - width + 1)
		top = rrow * (room_height + room_gap) + randrange(room_height - height + 1)
		right = left + width
		bottom = top + height
		self.left = left
		self.top = top
		self.right = right
		self.bottom = bottom
		positions = []
		for row, col in self.all_coords(0):
			cell = cells[row][col]
			cell.floor = floor
			if row == top or row == bottom - 1 or col == left or col == right - 1:
				cell.wall = wall
			else:
				positions.append((row, col))
			cell.room = self
		#print "Room: available positions =", positions ###
		shuffle(positions)
		self.available_positions = positions
	
	def __str__(self):
		return "room %s of level %s" % (self.coords, self.level.number)
	
	def room_closer_to_start(self):
		for corridor in self.corridors:
			room = corridor.other_room(self)
			if room.distance_from_start == self.distance_from_start:
				if debug:
					print >>db, "*** Cannot decide which room is closer to start:", room, self ###
			if room.distance_from_start < self.distance_from_start:
				return room
		return self
	
	def rebuild_available_positions(self):
		#print "Rebuilding available positions for room" ###
		positions = []
		for row, col in self.all_coords(1):
			cell = self.level.get_cell(row, col)
			if not (cell.stairs or cell.item):
				positions.append((row, col))
		#print "Available positions =", positions ###
		shuffle(positions)
		self.available_positions = positions
	
	def all_coords(self, pad):
		for row in xrange(self.top + pad, self.bottom - pad):
			for col in xrange(self.left + pad, self.right - pad):
				yield row, col
	
	def get_exit(self, drow, dcol):
		ecoords = self.exits.get((drow, dcol))
		if not ecoords:
			if drow:
				ecol = randrange(self.left + 1, self.right - 1)
				if drow < 0:
					erow = self.top
				else:
					erow = self.bottom - 1
			else:
				erow = randrange(self.top + 1, self.bottom - 1)
				if dcol < 0:
					ecol = self.left
				else:
					ecol = self.right - 1
			ecoords = erow, ecol
			self.exits[drow, dcol] = ecoords
		return ecoords
	
	def random_coords(self, pad = 0):
		row = randrange(self.top + 1 + pad, self.bottom - 1 - pad)
		col = randrange(self.left + 1 + pad, self.right - 1 - pad)
		return row, col
	
	def random_empty_position_for(self, item):
		if item.not_near_wall:
			pad = 1
		else:
			pad = 0
		return self.random_empty_position(pad)

	def random_empty_position(self, pad = 0):
		if debug:
			print >>db, "Getting random empty position in", self, "with pad", pad ###
			print >>db, "Room bounds are row", (self.top, self.bottom), "col", (self.left, self.right) ###
		d = pad + 1
		for coords in self.available_positions:
			row, col = coords
			if (self.left + d <= col < self.right - d and
					self.top + d <= row < self.bottom - d):
				self.available_positions.remove(coords)
				if debug:
					print >>db, "Found", coords ###
				return coords
		if debug:
			print >>db, "None found" ###
		return None
	
	def reveal(self, level):
		if not self.revealed:
			self.revealed = True
			for row in xrange(self.top, self.bottom):
				for col in xrange(self.left, self.right):
					level.get_cell(row, col).revealed = True

#-----------------------------------------------------------------------

class Corridor(object):
	coords = None
	rooms = None
	contains_obstacle = False
	
	def __init__(self):
		self.coords = []
	
	def other_room(self, room):
		rooms = self.rooms[:]
		rooms.remove(room)
		return rooms[0]
	
	def room_closer_to_start(self):
		room1, room2 = self.rooms
		if room1.distance_from_start == room2.distance_from_start:
			if debug:
				print >>db, "*** Cannot decide which room is closer to start:", room1, room2 ###
		if room1.distance_from_start < room2.distance_from_start:
			return room1
		else:
			return room2

#-----------------------------------------------------------------------

def make_path(level, path, thickness, proc, collect = None):
	for i in xrange(len(path) - 1):
		row1, col1 = path[i]
		row2, col2 = path[i+1]
		top = min(row1, row2)
		bottom = max(row1, row2)
		left = min(col1, col2)
		right = max(col1, col2)
		for row in xrange(top - thickness, bottom + thickness + 1):
			for col in xrange(left - thickness, right + thickness + 1):
				cell = level.get_cell(row, col)
				if cell:
					proc(cell)
					if collect is not None:
						coords = row, col
						if coords not in collect:
							collect.append(coords)

def make_corridor(level, path, floor, wall):
	def put_wall(cell):
		if not cell.floor:
			cell.floor = floor
			if not cell.wall:
				cell.wall = wall
	def put_space(cell):
		cell.wall = None
	corridor = Corridor()
	make_path(level, path, 1, put_wall)
	make_path(level, path, 0, put_space, corridor.coords)
	return corridor

def connect_rooms(level, room1, room2):
	rrow1, rcol1 = room1.coords
	rrow2, rcol2 = room2.coords
	drow = rrow2 - rrow1
	dcol = rcol2 - rcol1
	erow1, ecol1 = room1.get_exit(drow, dcol)
	erow2, ecol2 = room2.get_exit(-drow, -dcol)
	path = [(erow1, ecol1)]
	if erow1 <> erow2 and ecol1 <> ecol2:
		if drow:
			m = (erow1 + erow2) / 2
			path.extend([(m, ecol1), (m, ecol2)])
		else:
			m = (ecol1 + ecol2) / 2
			path.extend([(erow1, m), (erow2, m)])
	path.append((erow2, ecol2))
	corridor = make_corridor(level, path, choice(floors), choice(walls))
	room1.corridors.append(corridor)
	room2.corridors.append(corridor)
	corridor.rooms = [room1, room2]
	level.corridors.append(corridor)

def adjacent_unconnected_rooms(level, room):
	result = []
	rrow, rcol = room.coords
	for drow, dcol in directions:
		room2 = level.get_room(rrow + drow, rcol + dcol)
		if room2 and not room2.connected:
			result.append(room2)
	return result

def make_room_connections(level):
	room = level.rooms[0][0]
	room.connected = True
	candidates = [room]  #	Connected rooms which may have adjacent unconnected rooms
	while candidates:
		room = choice(candidates)
		adj = adjacent_unconnected_rooms(level, room)
		if adj:
			room2 = choice(adj)
			connect_rooms(level, room, room2)
			room2.connected = True
			candidates.append(room2)
		else:
			candidates.remove(room)

def populate_levels(levels, repopulating = False):
	placements = []
	for i, level in enumerate(levels):
		if repopulating:
			recipe = repopulation_recipe
		else:
			recipe = level_recipes[i]
		expand_recipe(recipe, level, placements)
	shuffle(placements)
	promote_items_on_through_path(placements)
	demote_items_placed_last(placements)
	group_prerequisites(placements)
	perform_placements(placements, reverse = repopulating)

def promote_items_on_through_path(placements):
	for i in xrange(len(placements)):
		if placements[i][1].on_through_path:
			#print "Moving", placements[i][1], "to beginning of placements" ###
			placements.insert(0, placements.pop(i))

def demote_items_placed_last(placements):
	last = []
	for placement in placements:
		if placement[1].place_last:
			last.append(placement)
	for placement in last:
		#print "Moving", placement[1], "to end of placements" ###
		placements.remove(placement)
		placements.append(placement)

def group_prerequisites(placements):
	last_item_with_class = {}
	for level, item in placements:
		if item.group_prerequisites:
			key = (level, item.__class__)
			if key in last_item_with_class:
				last_item = last_item_with_class[key]
				#print "Transferring prerequisites from", last_item, "to", item ###
				item.prerequisites.extend(last_item.prerequisites)
				del last_item.prerequisites[:]
			last_item_with_class[key] = item

def expand_recipe(recipe, level, placements, prerequisite_for = None):
	scale = (level.rooms_wide * level.rooms_high) / 9.0
	for cls, lo, hi in recipe:
		if not prerequisite_for and cls.scale_quantities:
			lo = int(round(scale * lo))
			hi = int(round(scale * hi))
		for j in xrange(randrange(lo, hi + 1)):
			item = cls(prerequisite_for)
			placements.append((level, item))
			if item.auxiliary_recipe:
				expand_recipe(item.auxiliary_recipe, level, item.prerequisites, prerequisite_for = item)

def perform_placements(placements, reverse):
	while placements:
		level, item = placements.pop(0)
		if debug:
			print >>db, "Performing placement of", item.def_name(), "on", level ###
		place_item(level, item, reverse)
		if item.prerequisites:
			for prerequisite in item.prerequisites:
				i = randrange(len(placements) + 1)
				placements.insert(i, prerequisite)
		if debug:
			print >>db, "Performed placement of", item.def_name() ###

def place_item(level, item, reverse):
	if item.prerequisite_for:
		place_prerequisite(item, reverse)
	elif item.is_prerequisite:
		place_in_room_reachable_from(item, level.entrance_room, item.level_jitter)
	else:
		place_regular_item(level, item)

def place_prerequisite(item, reverse):
	pf = item.prerequisite_for
	if debug:
		print >>db, "Placing prerequisite for", pf ###
	if pf.in_corridor:
		from_room = pf.corridor.room_closer_to_start()
	else:
		from_room = pf.get_room().room_closer_to_start()
	if not reverse:
		cross_level_chance = item.level_jitter
	else:
		cross_level_chance = 0
	place_in_room_reachable_from(item, from_room, cross_level_chance)

def place_in_room_reachable_from(item, from_room, cross_level_chance):
	for room in reachable_rooms(from_room, cross_level_chance):
		if place_item_in_room(item, room):
			return
	if debug:
		print >>db, "*** No position found for", item ###

def place_regular_item(level, item):
	while level.level_above and chance(item.level_jitter):
		level = level.level_above
	if item.in_corridor:
		place_item_in_corridor_on_level(level, item)
	else:	
		place_item_in_room_on_level(level, item)
	
def place_item_in_corridor_on_level(level, item):
	corridor, coords = level.random_corridor_position(item.on_through_path)
	corridor.contains_obstacle = True
	item.corridor = corridor
	place_item_on_level_at_coords(item, level, coords)
	#item.move_to_coords(row, col, level)
	
def place_item_in_room_on_level(level, item):
	roompos = level.random_room_position_for(item)
	if roompos:
		room, coords = roompos
		if item.is_obstacle:
			room.contains_obstacle = True
		place_item_in_room_at_coords(item, room, coords)
	else:
		if debug:
			print >>db, "*** No available position for", item ###

def place_item_in_room(item, room):
	coords = room.random_empty_position_for(item)
	if coords:
		place_item_in_room_at_coords(item, room, coords)
		return True

def place_item_in_room_at_coords(item, room, coords):
	if debug:
		print >>db, "Placing", item.def_name(), "in", room, "at", coords ###
	place_item_on_level_at_coords(item, room.level, coords)

def place_item_on_level_at_coords(item, level, coords):
	row, col = coords
	item.move_to_coords(row, col, level)

def reachable_rooms(from_room, cross_level_chance):
	if debug:
		print >>db, "Finding rooms reachable from room", from_room.coords, "of level", from_room.level.number ###
	visited = set()
	def traverse(room):
		if room not in visited:
			visited.add(room)
			if debug:
				print >>db, "Room", room.coords, "of level", room.level.number, "is a candidate" ###
			level = room.level
			if room is level.entrance_room and level.level_above and chance(cross_level_chance):
				for room2 in traverse(level.level_above.exit_room):
					yield room2
			if room is level.exit_room and level.level_below and chance(cross_level_chance):
				for room2 in traverse(level.level_below.entrance_room):
					yield room2
			corridors = room.corridors[:]
			shuffle(corridors)
			for corridor in corridors:
				room2 = corridor.other_room(room)
				if corridor.contains_obstacle or room2.contains_obstacle:
					if not room2.distance_from_start < room.distance_from_start:
						continue
				for room3 in traverse(room2):
					yield room3
			yield room
	return traverse(from_room)

def calculate_distances_from_start(room, d, reverse):
	#print "Level", room.level.number, "room", room.coords, "distance from start =", d ###
	room.distance_from_start = d
	for corridor in room.corridors:
		room2 = corridor.other_room(room)
		if room2.distance_from_start is None:
			calculate_distances_from_start(room2, d + 1, reverse)
		if reverse:
			if room2 is room2.level.entrance_room:
				level2 = room2.level.level_above
				if level2:
					calculate_distances_from_start(level2.exit_room, d + 1, reverse)
		else:
			if room2 is room2.level.exit_room:
				level2 = room2.level.level_below
				if level2:
					calculate_distances_from_start(level2.entrance_room, d + 1, reverse)

def repopulate_dungeon(levels):
	#print "Repopulating the dungeon" ###
	for level in levels:
		level.clear_distances_from_start()
		level.rebuild_available_positions()
	calculate_distances_from_start(levels[-1].exit_room, 0, reverse = True)
	populate_levels(levels[:-1], repopulating = True)

#-----------------------------------------------------------------------

class Level(object):
	game = None
	number = None
	cells = None
	rooms = None
	actors = None
	corridors = None
	entrance_coords = None
	exit_coords = None
	entrance_room = None
	exit_room = None
	level_above = None
	level_below = None
	path_through_level = None
	animated_items = None
	
	def __init__(self, game, number, level_above, (rooms_high, rooms_wide)):
		self.rooms_high = rooms_high
		self.rooms_wide = rooms_wide
		self.dungeon_cells_wide = rooms_wide * room_width + (rooms_wide - 1) * room_gap
		self.dungeon_cells_high = rooms_high * room_height + (rooms_high - 1) * room_gap
		col_range = xrange(self.dungeon_cells_wide)
		row_range = xrange(self.dungeon_cells_high)
		self.game = game
		self.number = number
		self.level_above = level_above
		if level_above:
			level_above.level_below = self
		self.actors = []
		self.animated_items = []
		cells = [[Cell() for col in col_range] for row in row_range]
		self.cells = cells
		self.corridors = []
		self.rooms = [[Room(cells, rrow, rcol, self)
			for rcol in xrange(rooms_wide)]
				for rrow in xrange(rooms_high)]
		make_room_connections(self)
		self.entrance_room, self.entrance_coords = self.add_stairs(upstairs)
		if self.number < len(level_recipes) - 1:
			exit_tile = downstairs
		else:
			exit_tile = None
		self.exit_room, self.exit_coords = self.add_stairs(exit_tile, exclude = self.entrance_room)
		self.path_through_level = self.find_path(self.entrance_room, self.exit_room)
		if debug:
			self.dump_level(db)
		#self.follow_recipe(level_recipes[number])
	
	def __str__(self):
		return "level %s" % self.number
	
	def dump_level(self, f):
		print >>f, "-" * 79
		print >>f, "Level", self.number
		entrance_cell = self.get_cell(*self.entrance_coords)
		exit_cell = self.get_cell(*self.exit_coords)
		for cell_row in self.cells:
			chars = []
			for cell in cell_row:
				if cell is entrance_cell:
					c = 'u'
				elif cell is exit_cell:
					c = 'd'
				elif cell.wall:
					c = '#'
				elif cell.floor:
					c = '.'
				else:
					c = ' '
				chars.append(c)
			print >>f, "".join(chars)
		print >>f, "-" * 79
	
	def clear_distances_from_start(self):
		for room in self.all_rooms():
			room.distance_from_start = None
	
	def add_grail(self):
		grail = Grail()
		row, col = self.exit_coords
		grail.move_to_coords(row, col, self)
	
	def add_stairs(self, tile, exclude = None):
		room = self.random_room(exclude)
		coords = room.random_empty_position(pad = 1)
		cell = self.get_cell(*coords)
		cell.stairs = tile
		return room, coords
	
	def find_path(self, from_room, to_room):
		path = []
		visited = set()
		def find_path_from(room):
			visited.add(room)
			if room is to_room:
				return True
			else:
				for corridor in room.corridors:
					room2 = corridor.other_room(room)
					if room2 not in visited:
						if find_path_from(room2):
							path.append(corridor)
							return True
				return False
		if not find_path_from(from_room):
			raise RuntimeError("No path from entrance to exit on level %d" % self.number)
		return path

	def follow_recipe(self, recipe, reverse = False):
		pending_recipes = [(recipe, False)]
		while pending_recipes:
			recipe, is_prerequisite = pending_recipes.pop()
			for (cls, min_number, max_number) in recipe:
				for i in xrange(randrange(min_number, max_number + 1)):
					#print "Placing an instance of", cls ###
					item = cls()
					self.add_item(item, is_prerequisite, reverse)
					aux = item.auxiliary_recipe
					if aux:
						pending_recipes.append((aux, True))
	
	def previous_level(self, reverse):
		if reverse:
			return self.level_below
		else:
			return self.level_above
	
	def get_cell(self, row, col):
		if 0 <= row < self.dungeon_cells_high and 0 <= col < self.dungeon_cells_wide:
			return self.cells[row][col]
		else:
			return 0

	def get_room(self, rrow, rcol):
		if 0 <= rrow < self.rooms_high and 0 <= rcol < self.rooms_wide:
			return self.rooms[rrow][rcol]
		else:
			return None
	
	def all_rooms(self):
		for rrow in self.rooms:
			for room in rrow:
				yield room
	
	def random_room(self, exclude = None):
		while 1:
			room = choice(choice(self.rooms))
			if room is not exclude:
				return room
	
	def random_position(self, pad = 0):
		for z in xrange(100):
			room = choice(choice(self.rooms))
			row, col = room.random_coords(pad)
			cell = self.get_cell(row, col)
			if not cell.item and not cell.stairs:
				return row, col
			
	def random_position_in_room(self, room, pad = 0):
		for z in xrange(100):
			row, col = room.random_coords(pad)
			cell = self.get_cell(row, col)
			if not cell.item and not cell.stairs:
				return row, col
	
	def random_room_position_for(self, item):
		if item.not_near_wall:
			pad = 1
		else:
			pad = 0
		return self.random_room_position(pad)

	def random_room_position(self, pad):
		candidates = list(self.all_rooms())
		shuffle(candidates)
		for room in candidates:
			coords = room.random_empty_position(pad)
			if coords:
				return room, coords
		return None
	
	def random_corridor_position(self, on_through_path = False):
		if on_through_path:
			corridors = self.path_through_level
		else:
			corridors = self.corridors
		candidates = corridors #[c for c in corridors if not c.contains_obstacle]
		if candidates:
			corridor = choice(candidates)
			return corridor, choice(corridor.coords)
		else:
			print "*** No corridor position on path through level %d" % self.number ###
			#raise DungeonGenerationError("No unblocked corridor on path through level %d" % self.number)
	
	def draw(self, surface):
		cells = self.cells
		row_min, col_min, row_max, col_max = window.viewport(self)
		row_range = xrange(row_min, row_max)
		col_range = xrange(col_min, col_max)
		for row in row_range:
			for col in col_range:
				cells[row][col].draw(surface, row, col)
	
	def do_animation(self):
		for item in self.animated_items:
			item.do_animation()

	def rebuild_available_positions(self):
		for corridor in self.corridors:
			corridor.contains_obstacle = False
		for room in self.all_rooms():
			room.contains_obstacle = False
			room.rebuild_available_positions()
	
	def show_map(self, surface, box):
		buf_width = self.dungeon_cells_wide * cell_width
		buf_height = self.dungeon_cells_high * cell_height
		buffer = Surface((buf_width, buf_height))
		buffer.fill(black)
		col_range = xrange(self.dungeon_cells_wide)
		row_range = xrange(self.dungeon_cells_high)
		cells = self.cells
		for row in row_range:
			for col in col_range:
				cells[row][col].draw(buffer, row, col, True)
		hscale = float(box.width) / buf_width
		vscale = float(box.height) / buf_height
		scale = min(hscale, vscale)
		surface.fill(black, box)
		#map = pygame.transform.scale(buffer, (int(buf_width * scale), int(buf_height * scale)))
		map = pygame.transform.rotozoom(buffer, 0, scale)
		dest = map.get_rect()
		dest.center = box.center
		surface.blit(map, dest)
		pygame.display.flip()
		get_key()

#-----------------------------------------------------------------------

def load_player_images():
	right = load("player.png")
	left = pygame.transform.flip(right, 1, 0)
	up = pygame.transform.rotate(right, 90)
	down = pygame.transform.flip(up, 0, 1)
	return {
		DIR_RIGHT: right,
		DIR_DOWN: down,
		DIR_LEFT: left,
		DIR_UP: up,
	}

#-----------------------------------------------------------------------

class Player(Item):
	images = load_player_images()
	show_on_map = True
	level = None
	position = None
	facing = DIR_RIGHT
	health = 10
	max_health = 10
	fortitude = 5
	base_fortitude = 5
	intelligence = 5
	moral_fibre = 5
	base_moral_fibre = 5
	general_knowledge = 0
	inventory = None
	death_message = None
	has_grail = False
	won = False
	lost = False
	defeated = False
	
	def __init__(self, level, (row, col)):
		self.inventory = []
		self.level = level
		self.move_to_coords(row, col)
	
	def draw(self, surface, row, col, map):
		draw_tile(surface, self.images[self.facing], row, col, map)
	
	def move(self, (drow, dcol), is_repeat):
		row, col = self.coords
		row2 = row + drow
		col2 = col + dcol
		cell = self.level.get_cell(row2, col2)
		if not cell or cell.wall:
			raise AbortCommand
		item = cell.item
		if item:
			item.make_seen()
			if is_repeat:
				raise AbortCommand
			item.attack(self)
		else:
			self.step_to_coords(row2, col2)
			self.rest()
	
	def step_to_coords(self, row, col):
		self.move_to_coords(row, col)
		cell = self.level.get_cell(row, col)
		if cell and cell.stairs:
			self.use_stairs(cell.stairs)
	
	def use_stairs(self, stairs):
		level = self.level
		if stairs is upstairs:
			action = "ascend"
			new_level = level.level_above
			if new_level:
				row, col = new_level.exit_coords
		else:
			action = "descend"
			new_level = level.level_below
			row, col = new_level.entrance_coords
		if new_level:
			log("You %s the stairs." % action)
			self.move_to_coords(row, col, new_level)
		else:
			if self.has_grail:
				log("You ascend the stairs... to victory.")
				self.won = True
			else:
				log("You can't leave now, you don't have the grail!")

	def move_to_coords(self, *args):
		Item.move_to_coords(self, *args)
		self.scroll_window()
	
	def scroll_window(self):
		row, col = self.coords
		cell = self.level.get_cell(row, col)
		room = cell.room
		if isinstance(room, Room):
			window.reveal_rect(room.left, room.top, room.right, room.bottom)
		window.reveal_rect(col - 2, row - 2, col + 3, row + 3)

	def add_to_inventory(self, item):
		self.inventory.append(item)
	
	def remove_from_inventory(self, item):
		self.inventory.remove(item)
	
	def add_health(self, x):
		#self.health = max(0, min(self.health + x, self.max_health))
		self.health = max(0, self.health + x)
	
	def add_fortitude(self, x):
		self.fortitude = max(0, self.fortitude + x)
	
	def add_intelligence(self, x):
		self.intelligence += x
	
	def add_moral_fibre(self, x):
		self.moral_fibre = max(0, self.moral_fibre + x)
	
	def add_general_knowledge(self, x):
		self.general_knowledge += x
	
	def restore_health(self):
		self.health = self.max_health
		self.fortitude = self.base_fortitude
		self.moral_fibre = self.base_moral_fibre
		self.defeated = False
	
	def move_to_entrance(self):
		if self.has_grail:
			coords = self.level.exit_coords
		else:
			coords = self.level.entrance_coords
		self.move_to_coords(*coords)

	def take_damage(self, damage, actor):
		self.add_health(-damage)
		if self.health <= 0:
			self.die(actor)
	
	def die(self, actor):
		self.defeated = True
		#fatal = True
		fatal = not debug
		spam = [x for x in self.inventory if isinstance(x, Spam)]
		if len(spam) >= 2:
			fatal = False
			for x in spam[:2]:
				self.remove_from_inventory(x)
		self.death_message = actor.get_victory_message(fatal)
		self.revival_message = actor.get_revival_message()
		self.lost = fatal
	
	def rest(self):
		if self.health < self.max_health:
			self.health += 1
		if self.fortitude < self.base_fortitude:
			self.fortitude += 1
		if self.moral_fibre < self.base_moral_fibre:
			self.moral_fibre += 1
	
	def pass_time(self):
		#print "Player.pass_time" ###
		if self.health > self.max_health:
			self.health -= 1
		if self.fortitude > self.base_fortitude:
			self.fortitude -= 1
		if self.moral_fibre > self.base_moral_fibre:
			self.moral_fibre -= 1
		for actor in self.level.actors:
			actor.take_turn(self)
	
	def get_target_item(self):
		row, col = self.coords
		drow, dcol = self.facing
		cell = self.level.get_cell(row + drow, col + dcol)
		if cell:
			return cell.item
		else:
			return None
	
	def unconscious(self):
		if self.defeated and not self.lost:
			#print "Unconscious" ###
			return True
		else:
			#print "Not unconscious" ###
			return False
	
	def nearby_position(self):
		row, col = self.coords
		for drow in xrange(-1, 2):
			for dcol in xrange(-1, 2):
				if drow or dcol:
					cell = self.level.get_cell(row + drow, col + dcol)
					if cell and not cell.wall and not cell.item and not cell.stairs:
						return row + drow, col + dcol
		return None

#-----------------------------------------------------------------------

key_to_direction = {
	KEY_RIGHT: DIR_RIGHT,
	KEY_DOWN: DIR_DOWN,
	KEY_LEFT: DIR_LEFT,
	KEY_UP: DIR_UP,
}

#-----------------------------------------------------------------------

class Game(object):
	screen = None
	player = None
	levels = None
	actors = None
	last_key_event = None
	repeat_counter = 0
	
	def __init__(self, screen, level_size):
		self.screen = screen
		self.actors = set()
		init_color_names()
		num_levels = len(level_recipes)
		level_numbers = xrange(num_levels)
		levels = []
		last_level = None
		for n in level_numbers:
			level = Level(self, n, last_level, level_size)
			levels.append(level)
			last_level = level
		calculate_distances_from_start(levels[0].entrance_room, 0, reverse = False)
		last_level.add_grail()
		populate_levels(levels)
		self.levels = levels
		level = levels[max(-num_levels, min(options.start_level, num_levels - 1))]
		self.player = Player(level, level.entrance_coords)
		self.player.game = self
		message_log.clear()
	
	def repopulate_dungeon(self):
		repopulate_dungeon(self.levels)
	
	def run(self):
		player = self.player
		while 1:
			self.update_visibility()
			#self.scroll_window()
			self.refresh()
			if player.won or player.lost:
				break
			if player.unconscious():
				self.display_second_chance()
			else:
				self.handle_one_event()
		if player.won:
			animation = self.display_victory()
		else:
			animation = self.display_defeat()
		return get_yorn(animation)
	
	def handle_one_event(self):
		event = pygame.event.wait()
		type = event.type
		if type == QUIT:
			sys.exit(0)
		elif type == KEYDOWN:
			self.last_key_event = event
			self.repeat_counter = repeat_delay
			return self.do_key_command(event, is_repeat = False)
		elif type == KEYUP:
			self.last_key_event = None
		elif type == USEREVENT:
			self.player.level.do_animation()
			event = self.last_key_event
			if event:
				self.repeat_counter -= 1
				if self.repeat_counter == 0:
					self.repeat_counter = repeat_spacing
					return self.do_key_command(event, is_repeat = True)
	
	def do_key_command(self, event, is_repeat):
		try:
			if not is_repeat:
				if self.do_escape_key(event):
					return
				elif self.do_map_key(event):
					return
			message_log.mark()
			if self.do_movement_key(event, is_repeat):
				self.pass_time()
				return
			if not is_repeat:
				if self.do_inventory_key(event) \
						or self.do_other_key(event):
					self.pass_time()
		except AbortCommand:
			pass
	
	def pass_time(self):
		self.player.pass_time()
		for actor in self.actors:
			actor.take_turn(self.player)
	
	def update_visibility(self):
		player = self.player
		level = player.level
		row, col = player.coords
		for drow in xrange(-1, 2):
			for dcol in xrange(-1, 2):
				cell = level.get_cell(row + drow, col + dcol)
				cell.revealed = True
				if not (drow or dcol):
					room = cell.room
					if room:
						room.reveal(level)
				if drow and not dcol or dcol and not drow:
					item = cell.item
					if item:
						item.make_seen()
	
	def display_victory(self):
		screen = self.screen
		box = Rect(0, 0, 500, 400)
		box.center = window.arena_rect.center
		screen.fill(win_bg_color, box)
		frame.draw(screen, box)
		draw_text(screen, "Congratulations!",
			congrat_font, win_text_color,
			(box.centerx, box.top + 60))
		draw_text(screen, "You escaped with the Holy Grail.",
			victory_font, win_text_color,
			(box.centerx, box.bottom - 70))
		draw_text(screen, "Play Again? (Y/N)",
			play_again_font, win_text_color,
			(box.centerx, box.bottom - 30))
		pygame.display.flip()
		return GrailAnimation(screen, box.centerx, box.centery, 8)
	
	def display_second_chance(self):
		#print "display_second_chance" ###
		player = self.player
		screen = self.screen
		box = Rect(0, 0, 520, 310)
		box.center = window.arena_rect.center
		screen.fill(sc_bg_color, box)
		frame.draw(screen, box)
		tc = TextCursor(screen, margin = box.centerx, top = box.top + 10,
			font = log_font, color = sc_text_color, just = 0)
		tc.writeln("Ugh!", font = sc_font)
		tc.y += 20
		tc.writeln(player.death_message)
		tc.y += 20
		message = textwrap.wrap(player.revival_message.replace("\t", ""),
			sc_wrap_width, expand_tabs = False, replace_whitespace = True)
		#print repr(message) ###
		for line in message:
			tc.writeln(line)
		tc.y += 20
		tc.writeln("Press ENTER to continue playing")
		pygame.display.flip()
		#print "...waiting for key" ###
		get_continue_or_quit()
		#print "...got key" ###
		player.restore_health()
		player.move_to_entrance()

	def display_defeat(self):
		screen = self.screen
		box = Rect(0, 0, 500, 300)
		box.center = window.arena_rect.center
		screen.fill(rip_bg_color, box)
		frame.draw(screen, box)
		draw_text(screen, "RIP", rip_font, rip_text_color,
			(box.centerx, box.centery - 50))
		draw_text(screen, self.player.death_message, log_font, rip_text_color,
			(box.centerx, box.bottom - 100))
		draw_text(screen, "Play Again? (Y/N)", play_again_font, rip_text_color,
			(box.centerx, box.bottom - 50))
		pygame.display.flip()
	
	def do_escape_key(self, event):
		if event.key == KEY_ESCAPE:
			show_instructions(self.screen, "continue playing")
	
	def do_map_key(self, event):
		if event.unicode.upper() == "M":
			self.player.level.show_map(self.screen, window.arena_rect)
			return True
	
	def do_movement_key(self, event, is_repeat):
		dir = key_to_direction.get(event.key)
		if dir:
			self.player.facing = dir
			self.player.move(dir, is_repeat)
			return True
	
	def do_inventory_key(self, event):
		key = event.unicode.upper()
		for item in self.player.inventory:
			if item.command == key:
				item.use(self.player)
				return True
	
	def do_other_key(self, event):
		if event.unicode == " ":
			self.player.rest()
			return True

	def refresh(self):
		screen = self.screen
		player = self.player
		level = player.level
		screen.fill(black, window.arena_rect)
		level.draw(screen)
		message_log.draw(screen)
		self.draw_status(player)
		pygame.display.flip()
	
	def draw_status(self, player):
		screen = self.screen
		screen.fill(status_bg_color, window.status_rect)
		margins = window.status_margins
		tc = TextCursor(self.screen, margin = margins.left, top = margins.top,
			color = status_text_color, font = stats_font)
		tc.font = stats_font
		tc.tabs = [0, -stats_position]
		tc.writeln("Level %d" % (player.level.number + 1))
		tc.y += 10
		tc2 = TextCursor(self.screen, margin = margins.right, top = tc.y,
			color = status_text_color, just = -1, font = stats_font)
		def writestat(label, value, base = None):
			value_color = status_text_color
			if base is not None:
				tc2.write("/%d" % base)
				if value <= base // 2:
					value_color = status_warning_color
			tc2.writeln("%d" % value, color = value_color)
			tc.writeln(label)
		writestat("Health", player.health, player.max_health)
		writestat("Fortitude", player.fortitude, player.base_fortitude)
		writestat("Moral Fibre", player.moral_fibre, player.base_moral_fibre)
		writestat("Numeracy", player.intelligence)
		writestat("General Knowledge", player.general_knowledge)
		tc.y += 10
		tc.writeln("Inventory")
		tc.y += 5
		#tc.margin += 5
		tc.x = tc.margin
		tc.font = inventory_font
		bags = {}
		for item in player.inventory:
			bags.setdefault(item.get_name(1, long = True), []).append(item)
		keys = bags.keys()
		keys.sort
		for key in keys:
			bag = bags[key]
			count = len(bag)
			item = bag[0]
			desc = "%d %s" % (count, item.get_name(count, long = True))
			if item.command:
				desc = "%s (%s)" % (desc, item.command)
			tc.writeln(desc)
			
#-----------------------------------------------------------------------

def init_display():
	global window, message_log
	#print pygame.display.Info() ###
	#print pygame.display.get_wm_info() ###
	#print pygame.display.list_modes() ###
	if not options.desktop_size:
		dw, dh = default_desktop_size
		for mw, mh in pygame.display.list_modes():
			if mw >= dw and mh >= dh:
				w, h = mw, mh
		#print "Using mode %sx%s" % (w, h) ###
		options.desktop_size = w, h
	w, h = options.desktop_size
	flags = 0
	if options.fullscreen:
		flags |= FULLSCREEN
	else:
		h -= 40
	window = Window(w, h)
	screen = pygame.display.set_mode(window.screen_rect.size, flags)
	pygame.display.set_caption("Quest")
	message_log = MessageLog(window.log_rect)
	return screen

def show_title(surface):
	rect = surface.get_rect()
	box = Rect(0, 0, *title_box_size)
	box.center = rect.center
	surface.fill(title_bg_color, box)
	frame.draw(surface, box)
	tc = TextCursor(surface, margin = box.centerx, top = box.top + 20, just = 0, color = title_text_color)
	tc.writeln("Quest", font = title_font_1)
	tc.writeln("for the", font = title_font_2)
	tc.writeln("Holy Grail", font = title_font_3)
	animation = GrailAnimation(surface, box.centerx, tc.y + 90, 5)
	tc.y += 180
	tc.writeln("Version %s" % VERSION, font = version_font)
	tc.y += 20
	tc.writeln("Press ENTER to continue", font = log_font)
	pygame.display.flip()
	get_continue_or_quit(animation)

def show_instructions(surface, prompt):
	rect = surface.get_rect()
	box = Rect(0, 0, *instructions_box_size)
	box.center = rect.center
	surface.fill(instructions_bg_color, box)
	frame.draw(surface, box)
	tc = TextCursor(surface, margin = box.centerx, top = box.top + 20, just = 0, color = instructions_text_color)
	tc.writeln("How to Play", font = instructions_title_font)
	tc.y += 30
	tc.font = instructions_font
	#for line in instructions:
	for para in instructions.replace("\t", "").split("\n\n"):
		for line in textwrap.wrap(para, instructions_wrap_width, replace_whitespace = True):
			tc.writeln(line)
		tc.y += 10
	tc.y += 30
	tc.writeln("Press ENTER to %s, Q to quit" % prompt)
	pygame.display.flip()
	get_continue_or_quit()

def show_options(surface):
	rect = surface.get_rect()
	box = Rect(0, 0, *options_box_size)
	box.center = rect.center
	surface.fill(instructions_bg_color, box)
	frame.draw(surface, box)
	tc = TextCursor(surface, margin = box.centerx, top = box.top + 20,
		just = 0, color = instructions_text_color, font = options_font)
	tc.writeln("Select Level Size", font = options_heading_font)
	tc.y += 20
	tc.just = 1
	tc.margin = tc.x = box.left + 60
	tc2 = TextCursor(surface, margin = tc.margin + 30, top = tc.y,
		just = 1, color = instructions_text_color, font = options_font)
	keys = dungeon_sizes.keys()
	keys.sort()
	for key in keys:
		name, (rows, cols) = dungeon_sizes[key]
		tc.writeln(key)
		tc2.writeln("%s (%s x %s)" % (name, cols, rows))
	tc.y += 10
	tc2.y += 10
	tc.writeln("Q")
	tc2.writeln("Quit")
	pygame.display.flip()
	while 1:
		key = get_key()
		if key == "Q":
			sys.exit(0)
		elif key in dungeon_sizes:
			return dungeon_sizes[key][1]

#-----------------------------------------------------------------------

def display_credits(screen):
	screen.fill(black, window.screen_rect)
	w, h = credits_box_size
	box = Rect(0, 0, w, h)
	box.center = window.screen_rect.center
	tc = TextCursor(screen, margin = box.centerx, top = box.top, color = credits_text_color, just = 0)
	tc.writeln("Credits", font = credits_heading_font)
	tc.y += 30
	for role, names in credits:
		tc.writeln(role, font = credits_role_font)
		for name in names:
			tc.writeln(name, font = credits_name_font)
		tc.y += 20
	pygame.display.flip()
	get_continue_or_quit()		

def main():
	if options.seed is not None:
		seed(options.seed)
	pygame.init()
	init_colors()
	screen = init_display()
	pygame.time.set_timer(USEREVENT, frame_time)
	try:
		show_title(screen)
		if not options.skip_intro:
			show_instructions(screen, "play")
		if options.level_size:
			level_size = dungeon_sizes[options.level_size][1]
		elif options.small_dungeon:
			level_size = (2, 3)
		else:
			level_size = show_options(screen)
		while 1:
			game = Game(screen, level_size)
			sys.stdout.flush()
			if not game.run():
				break
	except SystemExit:
		pass
	if not options.skip_intro:
		display_credits(screen)

main()
